/*
 *	organ.c
 *
 *	generate note with /dev/speaker
 */

#include "organ.h"
#include "term.h"

static char *strdup2 __P_((char *));
static int cmpkb __P_((CONST VOID_P, CONST VOID_P));
static int searchkey __P_((int));
static int dispch __P_((int));
static int organ __P_((VOID_A));
static int getkey __P_((char **));
static char *getstr __P_((char *));
static int loadruncom __P_((VOID_A));

static char *initseq = INITSEQ;
static kbtable keybind[] = {
	{'`', "<g->", {0, 0}},
	{'1', "<a->", {1, 0}},
	{'2', "<b->", {2, 0}},
	{'3', NULL, {3, 0}},
	{'4', "d-", {4, 0}},
	{'5', "e-", {5, 0}},
	{'6', NULL, {6, 0}},
	{'7', "g-", {7, 0}},
	{'8', "a-", {8, 0}},
	{'9', "b-", {9, 0}},
	{'0', NULL, {10, 0}},
	{'-', ">d-<", {11, 0}},
	{'=', ">e-<", {12, 0}},
	{'\b', NULL, {13, 0}},

	{'\t', "<g>", {0, 1}},
	{'q', "<a>", {1, 1}},
	{'w', "<b>", {2, 1}},
	{'e', "c", {3, 1}},
	{'r', "d", {4, 1}},
	{'t', "e", {5, 1}},
	{'y', "f", {6, 1}},
	{'u', "g", {7, 1}},
	{'i', "a", {8, 1}},
	{'o', "b", {9, 1}},
	{'p', ">c<", {10, 1}},
	{'[', ">d<", {11, 1}},
	{']', ">e<", {12, 1}},
	{'\\', ">f<", {13, 1}},

	{'a', "<<a->>", {1, 2}},
	{'s', "<<b->>", {2, 2}},
	{'d', NULL, {3, 2}},
	{'f', "<d->", {4, 2}},
	{'g', "<e->", {5, 2}},
	{'h', NULL, {6, 2}},
	{'j', "<g->", {7, 2}},
	{'k', "<a->", {8, 2}},
	{'l', "<b->", {9, 2}},
	{';', NULL, {10, 2}},
	{'\'', "d-", {11, 2}},
	{CR, "e-", {12, 2}},

	{'z', "<<a>>", {1, 3}},
	{'x', "<<b>>", {2, 3}},
	{'c', "<c>", {3, 3}},
	{'v', "<d>", {4, 3}},
	{'b', "<e>", {5, 3}},
	{'n', "<f>", {6, 3}},
	{'m', "<g>", {7, 3}},
	{',', "<a>", {8, 3}},
	{'.', "<b>", {9, 3}},
	{'/', "c", {10, 3}},

	{' ', NULL, {-1, -1}},
	{'!', NULL, {-1, -1}},
	{'"', NULL, {-1, -1}},
	{'#', NULL, {-1, -1}},
	{'$', NULL, {-1, -1}},
	{'%', NULL, {-1, -1}},
	{'&', NULL, {-1, -1}},
	{'(', NULL, {-1, -1}},
	{')', NULL, {-1, -1}},
	{'*', NULL, {-1, -1}},
	{'+', NULL, {-1, -1}},
	{':', NULL, {-1, -1}},
	{'<', NULL, {-1, -1}},
	{'>', NULL, {-1, -1}},
	{'?', NULL, {-1, -1}},
	{'@', NULL, {-1, -1}},
	{'^', NULL, {-1, -1}},
	{'_', NULL, {-1, -1}},
	{'{', NULL, {-1, -1}},
	{'|', NULL, {-1, -1}},
	{'}', NULL, {-1, -1}},
	{'~', NULL, {-1, -1}},
	{'\177', NULL, {-1, -1}},

	{ESC, "", {-1, -1}}
};
#define	KBSIZE	(sizeof(keybind) / sizeof(kbtable))


VOID error(str)
char *str;
{
	endterm();
	inittty(1);
	fputc('\007', stderr);
	perror(str ? str : "organ");
	exit(1);
}

static char *strdup2(s)
char *s;
{
	char *tmp;

	if (!s) return(NULL);
	if (!(tmp = (char *)malloc((size_t)strlen(s) + 1))) error(NULL);
	strcpy(tmp, s);
	return(tmp);
}

static int cmpkb(vp1, vp2)
CONST VOID_P vp1;
CONST VOID_P vp2;
{
	return(((kbtable *)(vp1)) -> key - ((kbtable *)(vp2)) -> key);
}

static int searchkey(c)
int c;
{
	int i, j, min, max;

	min = 0;
	max = KBSIZE - 1;
	i = min;
	while (min < max) {
		i = (max + min) / 2;
		if (!(j = (c - keybind[i].key))) return(i);
		else if (j < 0) max = i - 1;
		else min = i + 1;
	}
	if (min >=0 && min < KBSIZE && c == keybind[min].key) return(min);
	return(-1);
}

static int dispch(n)
int n;
{
	if (n < 0 || n >= KBSIZE
	|| keybind[n].pos[0] < 0 || keybind[n].pos[1] < 0)
		return(-1);
	if (!(keybind[n].seq)) {
#if	1
		return(-1);
#else
		if (keybind[n].key < ' ' || keybind[n].key >= '\177')
			return(0);
		locate(keybind[n].pos[0] * 5 + keybind[n].pos[1] * 2 + 2,
			keybind[n].pos[1] + 1);
		putch2(keybind[n].key);
#endif
	}
	else if (keybind[n].key >= ' ' && keybind[n].key < '\177') {
		locate(keybind[n].pos[0] * 5 + keybind[n].pos[1] * 2 + 1,
			keybind[n].pos[1] + 1);
		putch2('[');
		putch2(keybind[n].key);
		putch2(']');
	}
	else {
		locate(keybind[n].pos[0] * 5 + keybind[n].pos[1] * 2 ,
			keybind[n].pos[1] + 1);
		putch2('[');
		putch2('^');
		putch2((keybind[n].key + '@') & 0x7f);
		putch2(']');
	}
	return(0);
}

static int organ(VOID_A)
{
	int i, c, old, fd, count;

	putterms(t_clear);
	locate(0, 0);
	cputs2("  Push ESC to exit.");
	for (i = 0; i < KBSIZE; i++) dispch(i);

	i = old = -1;
	for (;;) {
		locate(0, 0);
		tflush();

		if (i >= 0) {
			if ((fd = open(SPKRDEV, O_WRONLY)) < 0) error(SPKRDEV);
			write(fd, initseq, strlen(initseq));
			write(fd, keybind[i].seq, strlen(keybind[i].seq));
			write(fd, "\n", 1);
			close(fd);
		}

		count = SENSEPERSEC / 10;
		do {
			c = kbhit2(1000000L / SENSEPERSEC);
			if (old >= 0 && !(--count)) {
				dispch(old);
				locate(0, 0);
				tflush();
				old = -1;
			}
		} while	(!c);

		c = getch2();
		if (old >= 0) dispch(old);
		if ((i = searchkey(c)) < 0 || !(keybind[i].seq)) i = -1;
		else if (!*(keybind[i].seq)) break;
		else {
			putterm(t_standout);
			dispch(i);
			putterm(end_standout);
		}

		old = i;
	}
	tflush();
	return(0);
}

static int getkey(sp)
char **sp;
{
	int c, q;

	q = 0;
	c = *(*sp)++;
	if (c == '\'' && **sp) {
		if ((**sp != ' ' &&  **sp != '\t') || *(*sp + 1) == '\'') {
			c = *(*sp)++;
			q = '\'';
		}
	}
	if (c == '^' && **sp && **sp != ' ' && **sp != '\t' && **sp != q) {
		c = *(*sp)++;
		if (c >= 'a' && c <= 'z') c -= 'a' - 'A';
		if (c < '?' || c > '_') c = '\0';
		else c = ((c - '@') & 0x7f);
	}
	if (q) {
		if (**sp == '\'') (*sp)++;
		else c = '\0';
	}
	if (**sp && **sp != ' ' && **sp != '\t') c = '\0';
	while (**sp == ' ' || **sp == '\t') (*sp)++;
	return((c != ESC) ? c : '\0');
}

static char *getstr(s)
char *s;
{
	char *cp;

	if (*s == '"') {
		for (cp = ++s; (cp = strchr(cp, '"')); cp++) {
			if (cp <= s || *(cp - 1) != '\\') {
				*cp = '\0';
				break;
			}
		}
	}
	else {
		if ((cp = strpbrk(s, ";#"))) *cp = '\0';
		if (!*s) return(NULL);
	}
	return(strdup2(s));
}

static int loadruncom(VOID_A)
{
	struct passwd *pwd;
	FILE *fp;
	char *cp, buf[MAXBUF + 1];
	int i, j, c;

	if (!(cp = (char *)getenv("HOME"))) {
		if (!(pwd = getpwuid(getuid()))) return(-1);
		cp = pwd -> pw_dir;
	}
	if (chdir(cp) || !(fp = fopen(RCFILE, "r"))) return(-1);

	while (fgets(buf, MAXBUF, fp)) {
		for (cp = buf; *cp == ' ' || *cp == '\t'; cp++);
		if (!*cp || *cp == '#' || *cp == ';') continue;
		i = strlen(cp) - 1;
		if (cp[i] == '\n') cp[i] = '\0';

		for (i = 0; cp[i] && cp[i] != ' ' && cp[i] != '\t'; i++);
		for (j = i + 1; cp[j] == ' ' || cp[j] == '\t'; j++);
		if (i == sizeof(ID_INIT) - 1
		&& !strncasecmp(cp, ID_INIT, i)) {
			if (initseq) free(initseq);
			if ((cp = getstr(&(cp[j])))) initseq = cp;
		}
		else if (i == sizeof(ID_COPY) - 1
		&& !strncasecmp(cp, ID_COPY, i)) {
			cp += j;
			i = getkey(&cp);
			j = getkey(&cp);
			if (i && j
			&& (i = searchkey(i)) >= 0
			&& (j = searchkey(j)) >= 0) {
				if (keybind[i].seq) free(keybind[i].seq);
				keybind[i].seq = strdup2(keybind[j].seq);
			}
		}
		else if (i == sizeof(ID_SWAP) - 1
		&& !strncasecmp(cp, ID_SWAP, i)) {
			cp += j;
			i = getkey(&cp);
			j = getkey(&cp);
			if (i && j
			&& (i = searchkey(i)) >= 0
			&& (j = searchkey(j)) >= 0) {
				cp = keybind[i].seq;
				keybind[i].seq = keybind[j].seq;
				keybind[j].seq = cp;
				c = keybind[i].pos[0];
				keybind[i].pos[0] = keybind[j].pos[0];
				keybind[j].pos[0] = c;
				c = keybind[i].pos[1];
				keybind[i].pos[1] = keybind[j].pos[1];
				keybind[j].pos[1] = c;
			}
		}
		else if (i == sizeof(ID_CLR) - 1
		&& !strncasecmp(cp, ID_CLR, i)) {
			for (i = 0; i < KBSIZE; i++)
			if (keybind[i].key != ESC) {
				if (keybind[i].seq) {
					free(keybind[i].seq);
					keybind[i].seq = NULL;
				}
				keybind[i].pos[0] = keybind[i].pos[1] = -1;
			}
		}
		else if (i == sizeof(ID_NOTE) - 1
		&& !strncasecmp(cp, ID_NOTE, i)) {
			cp += j;
			i = getkey(&cp);
			cp = getstr(cp);
			if (i && (i = searchkey(i)) >= 0) {
				if (keybind[i].seq) free(keybind[i].seq);
				keybind[i].seq = cp;
			}
			else if (cp) free(cp);
		}
		else if (i == sizeof(ID_POS) - 1
		&& !strncasecmp(cp, ID_POS, i)) {
			cp += j;
			i = getkey(&cp);
			j = 0;
			while (cp[j] && cp[j] != ' ' && cp[j] != '\t') j++;
			while (cp[j] == ' ' || cp[j] == '\t') j++;
			if (i && *cp && cp[j] && (i = searchkey(i)) >= 0) {
				keybind[i].pos[0] = atoi(cp) - 1;
				keybind[i].pos[1] = atoi(&(cp[j])) - 1;
			}
		}
	}
	fclose(fp);

	return(0);
}

int main(argc, argv)
int argc;
char *argv[];
{
	int i;

	initseq = strdup2(initseq);
	for (i = 0; i < KBSIZE; i++)
		if (keybind[i].seq) keybind[i].seq = strdup2(keybind[i].seq);
	qsort(keybind, KBSIZE, sizeof(kbtable), cmpkb);
	loadruncom();

	if ((i = open(SPKRDEV, O_WRONLY)) < 0) error(SPKRDEV);
	close(i);

	inittty(0);
	raw2();
	noecho2();
	nonl2();
	notabs();
	getterment();
	initterm();
	getwsize(80, 6);

	organ();

	locate(0, 4);
	cputs2("\r\n");
	tflush();
	exit2(0);
	return(0);
}
