/*
 *	term.c
 *
 *	Terminal Module
 */

#include "organ.h"

#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>

#include <termios.h>
typedef	struct termios	termioctl_t;
#define	tioctl(d, r, a)	((r) ? tcsetattr(d, (r) - 1, a) : tcgetattr(d, a))
#define	getspeed(t)	cfgetospeed(&t)
#define	REQGETP		0
#define	REQSETP		TCSAFLUSH + 1

#include <stdlib.h>
#include "term.h"

#define	MAXPRINTBUF	255
#define	GETSIZE		"\033[6n"
#define	SIZEFMT		"\033[%d;%dR"

extern int tgetent __P_((char *, char *));
extern int tgetnum __P_((char *));
extern int tgetflag __P_((char *));
extern char *tgetstr __P_((char *, char **));
extern char *tgoto __P_((char *, int, int));
extern int tputs __P_((char *, int, int (*)__P_((int))));

#define	BUFUNIT		16
#define	TERMCAPSIZE	2048

#define	STDIN		0
#define	STDOUT		1
#define	STDERR		2
#define	TTYNAME		"/dev/tty"

#define	not(x)		(~(x) & 0xffff)

static int err2 __P_((char *));
static int defaultterm __P_((VOID_A));
static int maxlocate __P_((VOID_A));
static int getxy __P_((int *, int *));
static int ttymode __P_((int, u_short, u_short, u_short, u_short,
		u_short, u_short, int, int));
static int tgetstr2 __P_((char **, char *));

short ospeed = 0;
char PC ='\0';
char *BC = NULL;
char *UP = NULL;
int n_column = 0;
int n_line = 0;
char *t_init = NULL;
char *t_end = NULL;
char *t_scroll = NULL;
char *t_keypad = NULL;
char *t_nokeypad = NULL;
char *t_clear = NULL;
char *t_normal = NULL;
char *t_standout = NULL;
char *end_standout = NULL;
char *c_locate = NULL;

static int ttyio;
static FILE *ttyout;

static int termflags;
#define	F_INITTTY	001
#define	F_TERMENT	002
#define	F_INITTERM	004


int inittty(reset)
int reset;
{
	static termioctl_t dupttyio;
	termioctl_t tty;

	if (reset && !(termflags & F_INITTTY)) return(0);
	if (tioctl(ttyio, REQGETP, &tty) < 0) {
		termflags &= ~F_INITTTY;
		err2(NULL);
	}
	if (!reset) {
		memcpy(&dupttyio, &tty, sizeof(termioctl_t));
		ospeed = getspeed(dupttyio);
		termflags |= F_INITTTY;
	}
	else if (tioctl(ttyio, REQSETP, &dupttyio) < 0) {
		termflags &= ~F_INITTTY;
		err2(NULL);
	}

	return(0);
}

static int ttymode(d, set, reset, iset, ireset, oset, oreset, min, time)
int d;
u_short set, reset, iset, ireset, oset, oreset;
int min, time;
{
	termioctl_t tty;

	if (tioctl(d, REQGETP, &tty) < 0) err2(NULL);
	if (set) tty.c_lflag |= set;
	if (reset) tty.c_lflag &= reset;
	if (iset) tty.c_iflag |= iset;
	if (ireset) tty.c_iflag &= ireset;
	if (oset) tty.c_oflag |= oset;
	if (oreset) tty.c_oflag &= oreset;
	if (min) {
		tty.c_cc[VMIN] = min;
		tty.c_cc[VTIME] = time;
	}
	if (tioctl(d, REQSETP, &tty) < 0) err2(NULL);
	return(0);
}

int raw2(VOID_A)
{
	ttymode(ttyio, 0, not(ISIG | ICANON | IEXTEN),
		IGNBRK, not(BRKINT | IXON), 0, not(OPOST), 1, 0);
	return(0);
}

int noecho2(VOID_A)
{
	ttymode(ttyio, 0, not(ECHO | ECHOE | ECHOK | ECHONL), 0, 0,
		0, 0, 0, 0);
	return(0);
}

int nonl2(VOID_A)
{
	ttymode(ttyio, 0, 0, 0, not(ICRNL), 0, not(ONLCR), 0, 0);
	return(0);
}

int notabs(VOID_A)
{
	ttymode(ttyio, 0, 0, 0, 0, OXTABS, 0, 0, 0);
	return(0);
}

int keyflush(VOID_A)
{
	tcflush(ttyio, TCIFLUSH);
	return(0);
}

int exit2(no)
int no;
{
	if (termflags & F_TERMENT) putterm(t_normal);
	endterm();
	inittty(1);
	keyflush();
	exit(no);
	return(0);
}

static int err2(mes)
char *mes;
{
	if (termflags & F_TERMENT) putterm(t_normal);
	endterm();
	inittty(1);
	fputs("\007\n", stderr);
	if (!mes) perror(TTYNAME);
	else {
		fputs(mes, stderr);
		fputc('\n', stderr);
	}
	exit(1);
	return(0);
}

static int defaultterm(VOID_A)
{
	int i;

	BC = "\010";
	UP = "\033[A";
	n_column = 80;
	n_line = 24;
	t_init = "";
	t_end = "";
	t_scroll = "\033[%i%d;%dr";
	t_keypad = "\033[?1h\033=";
	t_nokeypad = "\033[?1l\033>";
	t_clear = "\033[;H\033[2J";
	t_normal = "\033[m";
	t_standout = "\033[7m";
	end_standout = "\033[m";
	c_locate = "\033[%i%d;%dH";

	return(0);
}

static int maxlocate(VOID_A)
{
	locate(998, 998);
	return(0);
}

static int getxy(yp, xp)
int *yp, *xp;
{
	char *format, buf[sizeof(SIZEFMT) + 4];
	int i, j, k, tmp, count, *val[2];

	format = SIZEFMT;
	write(ttyio, GETSIZE, sizeof(GETSIZE) - 1);

	for (i = 0; i < sizeof(buf) - 1; i++) {
		buf[i] = getch2();
		if (buf[i] == format[sizeof(SIZEFMT) - 2]) break;
	}
	buf[++i] = '\0';

	count = 0;
	val[0] = yp;
	val[1] = xp;

	for (i = j = 0; format[i] && buf[j]; i++) {
		if (format[i] == '%' && format[++i] == 'd' && count < 2) {
			tmp = 0;
			k = j;
			for (; buf[j]; j++) {
				if (!buf[j] || buf[j] == format[i + 1]) break;
				tmp = tmp * 10 + buf[j] - '0';
			}
			if (j == k) break;
			*val[count++] = tmp;
		}
		else if (format[i] != buf[j++]) break;
	}
	return(count);
}

char *tparamstr(str, arg1, arg2)
char *str;
int arg1, arg2;
{
	char *buf;
	int i, j, n, s, size, args[2];

	if (!str) return(NULL);
	if (!(buf = (char *)malloc(size = BUFUNIT))) err2(NULL);
	args[0] = arg1;
	args[1] = arg2;

	for (i = j = n = 0; str[i]; i++) {
		if (j + 5 >= size) {
			size += BUFUNIT;
			if (!(buf = (char *)realloc(buf, size))) err2(NULL);
		}
		if (str[i] != '%') buf[j++] = str[i];
		else if (str[++i] == '%') buf[j++] = '%';
		else if (n >= 2) {
			free(buf);
			return(NULL);
		}
		else switch (str[i]) {
			case 'd':
				sprintf(buf + j, "%d", args[n++]);
				j += strlen(buf + j);
				break;
			case '2':
				sprintf(buf + j, "%2d", args[n++]);
				j += 2;
				break;
			case '3':
				sprintf(buf + j, "%3d", args[n++]);
				j += 3;
				break;
			case '.':
				sprintf(buf + (j++), "%c", args[n++]);
				break;
			case '+':
				if (!str[++i]) {
					free(buf);
					return(NULL);
				}
				sprintf(buf + (j++), "%c", args[n++] + str[i]);
				break;
			case '>':
				if (!str[++i] || !str[i + 1]) {
					free(buf);
					return(NULL);
				}
				if (args[n] > str[i++]) args[n] += str[i];
				break;
			case 'r':
				s = args[0];
				args[0] = args[1];
				args[1] = s;
				break;
			case 'i':
				args[0]++;
				args[1]++;
				break;
			case 'n':
				args[0] ^= 0140;
				args[1] ^= 0140;
				break;
			case 'B':
				args[n] = ((args[n] / 10) << 4)
					| (args[n] % 10);
				break;
			case 'D':
				args[n] -= 2 * (args[n] % 16);
				break;
			default:
				free(buf);
				return(NULL);
/*NOTREACHED*/
				break;
		}
	}

	buf[j] = '\0';
	return(buf);
}

static int tgetstr2(term, str)
char **term;
char *str;
{
	char strbuf[TERMCAPSIZE];
	char *p, *cp;

	p = strbuf;
	if ((cp = tgetstr(str, &p)) || (cp = *term)) {
		if (!(*term = (char *)malloc(strlen(cp) + 1))) err2(NULL);
		strcpy(*term, cp);
	}
	return(0);
}

int getterment(VOID_A)
{
	char *cp, *termname, buf[TERMCAPSIZE];
	int i;

	if (termflags & F_TERMENT) return(-1);
	if (!(termname = (char *)getenv("TERM"))) termname = "unknown";
	if (tgetent(buf, termname) <= 0) err2("No TERMCAP prepared");

	if ((ttyio = open(TTYNAME, O_RDWR, 0600)) < 0) ttyio = STDERR;
	if (!(ttyout = fopen(TTYNAME, "w+"))) ttyout = stdout;

	defaultterm();
	cp = "";
	tgetstr2(&cp, "pc");
	PC = *cp;
	free(cp);
	tgetstr2(&BC, "bc");
	tgetstr2(&UP, "up");

	n_column = tgetnum("co");
	n_line = tgetnum("li");
	tgetstr2(&t_init, "ti");
	tgetstr2(&t_end, "te");
	tgetstr2(&t_scroll, "cs");
	tgetstr2(&t_keypad, "ks");
	tgetstr2(&t_nokeypad, "ke");
	tgetstr2(&t_clear, "cl");
	tgetstr2(&t_normal, "me");
	tgetstr2(&t_standout, "so");
	tgetstr2(&end_standout, "se");
	tgetstr2(&c_locate, "cm");

	termflags |= F_TERMENT;
	return(0);
}

int initterm(VOID_A)
{
	if (!(termflags & F_TERMENT)) getterment();
	putterms(t_keypad);
	putterms(t_init);
	tflush();
	termflags |= F_INITTERM;
	return(0);
}

int endterm(VOID_A)
{
	if (!(termflags & F_INITTERM)) return(-1);
	putterms(t_nokeypad);
	putterms(t_end);
	tflush();
	termflags &= ~F_INITTERM;
	return(0);
}

int putch2(c)
int c;
{
	return(fputc(c, ttyout));
}

int cputs2(str)
char *str;
{
	return(fputs(str, ttyout));
}

int kbhit2(usec)
u_long usec;
{
	fd_set readfds;
	struct timeval timeout;

	timeout.tv_sec = 0L;
	timeout.tv_usec = usec;
	FD_ZERO(&readfds);
	FD_SET(STDIN, &readfds);

	return(select(1, &readfds, NULL, NULL, &timeout));
}

int getch2(VOID_A)
{
	u_char ch;
	int i;

	while ((i = read(ttyio, &ch, sizeof(u_char))) < 0 && errno == EINTR);
	if (i < sizeof(u_char)) return(EOF);
	return((int)ch);
}

int setscroll(s, e)
int s, e;
{
	char *cp;

	if ((cp = tparamstr(t_scroll, s, e))) {
		putterms(cp);
		free(cp);
	}
	return(0);
}

int locate(x, y)
int x, y;
{
	putterms(tgoto(c_locate, x, y));
	return(0);
}

int tflush(VOID_A)
{
	fflush(ttyout);
	return(0);
}

int getwsize(xmax, ymax)
int xmax, ymax;
{
	int x, y;
	struct winsize ws;

	x = y = -1;
	if (ioctl(ttyio, TIOCGWINSZ, &ws) >= 0) {
		if (ws.ws_col > 0) x = ws.ws_col;
		if (ws.ws_row > 0) y = ws.ws_row;
	}

	if (x < 0 || y < 0) {
		setscroll(-1, -1);
		maxlocate();
		if (getxy(&y, &x) != 2) x = y = -1;
	}

	if (x > 0) n_column = x;
	if (y > 0) n_line = y;

	if (n_column < xmax) err2("Column size too small");
	if (n_line < ymax) err2("Line size too small");

	setscroll(-1, n_line - 1);
	return(0);
}
